#include "hash.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "../gpmi.h"
#include "mods-static.h"
#include "pkgs-static.h"
#include "threads.h"
#include "os_dep.h"

void (*gpmi_debug_proc)(int level, char *s) = NULL;

/* These ones is safe to keep a global out of threads.h, as these are used only
when static packages are initialized, before main(), so no threading for sure.
*/
int gpmi_initialize_paths = 1;  /* Do not use it yourself, the doc explains how it should be used */
int gpmi_initialized      = 0;  /* Is gpmi initialised? */

/**
 * Main gpmi_init proc, which should ALWAYS be called!
 */
void gpmi_init() {

	/* Init threading - we need to do this before anything else because
	   we store some data in structures gpmi_threads_init will set up. */
	gpmi_threads_init();

	/* By default (when the app calls gpmi_init()) this part will run to set
	   up the paths if the app haven't set them up yet. Static packages
		 also run gpmi_init(), and they don't want paths to be initialized,
		 so they set gpmi_initialize_paths to 0 before the call. */
	if (gpmi_initialize_paths) {
		gpmi_shared_lock;
		/* Check if one of the paths aren't assigned, and do so if needed. */
		if (gpmi_shared.path_modules == NULL)
			gpmi_path_append(GPMI_PATH_MODULES, GPMI_PATH_DEFAULT);

		if (gpmi_shared.path_packages == NULL)
			gpmi_path_append(GPMI_PATH_PACKAGES, GPMI_PATH_DEFAULT);

		if (gpmi_shared.path_scripts == NULL)
			gpmi_path_append(GPMI_PATH_SCRIPTS, GPMI_PATH_DEFAULT);
		gpmi_shared_unlock;
	}

	/* Do not reinitialize */
	if (gpmi_initialized) {
		gpmi_initialize_paths = 1;
		return;
	}


	/* Init the keyword system so keyword tables are collected in a table
	 * TODO - This is optional. We may want to make a "light" option which
	 * disables this call. (keep the uninit in mind too!)
	*/
	gpmi_keyword_init();

	/* Init the event-system */
	/* TODO - Maybe this should be optional too? (keep the uninit in mind too!) */
	gpmi_event_init();

	/* Check for needed auto-load-packages if we already have paths */
	if (gpmi_initialize_paths)
		gpmi_libdep_load(&gpmi_shared.main_lib, "gpmi");

	gpmi_initialized      = 1;
	gpmi_initialize_paths = 1;
}

void gpmi_uninit()
{
	gpmi_package *package;

	/* Do not run uninit as long as there are static package loaded in the system */
	if (gpmi_shared.static_packages != NULL)
		return;

	/* Do not run uninit as long as there are static modules loaded in the system */
	if (gpmi_shared.static_modules != NULL)
		return;

	/* Unload any possible deps of the application */
	gpmi_libdep_unload(&gpmi_shared.main_lib);

	while (gpmi_shared.dynamic_packages != NULL) {
		gpmi_shared_lock;
		package = gpmi_shared.dynamic_packages;
		gpmi_shared.dynamic_packages = gpmi_shared.dynamic_packages->next;
		gpmi_shared_unlock;

		gpmi_pkg_unload(package);
	}

	/* Uninit the event-system */
	gpmi_event_uninit();

	/* Uninit the keyword-system */
	gpmi_keyword_uninit();

	/* Free paths if needed */
	if (gpmi_shared.path_modules != NULL) {
		free(gpmi_shared.path_modules);
		gpmi_shared.path_modules = NULL;
	}

	if (gpmi_shared.path_packages != NULL) {
		free(gpmi_shared.path_packages);
		gpmi_shared.path_packages = NULL;
	}

	if (gpmi_shared.path_scripts != NULL) {
		free(gpmi_shared.path_scripts);
		gpmi_shared.path_scripts = NULL;
	}

	gpmi_threads_uninit();

	gpmi_initialized = 0;
}
