#include "php.h"

#include <stdio.h>
#include <string.h>
#include <main/php_embed.h>

#include "gpmi/debug.h"

#define GPMI_PHP_MODULE_VERSION "1.0"

static int php_module_loaded = 0;

static zend_module_entry php_gpmi_module_entry = {
	STANDARD_MODULE_HEADER,
	"gpmi",
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	GPMI_PHP_MODULE_VERSION,
	STANDARD_MODULE_PROPERTIES
};

static int php_set_ini_entry(const char *entry, const char *value, int stage)
{
	int retval;

	retval = zend_alter_ini_entry((char *)entry, strlen(entry) + 1, (char *)value, strlen(value) + 1, PHP_INI_USER, stage);
	if (retval != SUCCESS) {
		gpmi_error(0, "[PHP] Failed setting INI entry %s to %s (stage = %d)", entry, value, stage);
	}

	return retval;
}

ZEND_NAMED_FUNCTION(PHP_DynamicPackageCommand)
{
	char *(*proc)(int argc, char *argv[]);
	pval ***php_args;
	char **args;
	char *res;
	int argc, i;

	proc = GPMI_Script_FindProc(get_active_function_name());

	if (proc == NULL) {
		gpmi_error(0, "[PHP] Internal error: %s has a NULL proc in the command hash\n", get_active_function_name());
		RETVAL_NULL();
		return;
	}

	argc = ZEND_NUM_ARGS();

	/* Allocate memory for the args-array */
	args = malloc((argc + 1) * sizeof(char *));
	php_args = malloc(argc * sizeof(pval **));

	/* Read params to an array */
	if (zend_get_parameters_array_ex(argc, php_args) == FAILURE) {
		gpmi_debug(0, "[PHP] Failed executing command");
		RETVAL_NULL();
		return;
	}

	args[0] = get_active_function_name();

	/* Convert everything to strings */
	for (i = 0; i < argc; i++) {
		convert_to_string_ex(php_args[i]);
		args[i + 1] = (*php_args[i])->value.str.val;
	}

	res = proc(argc + 1, args);

	/* Free the allocated memory */
	free(args);
	free(php_args);

	if (res == NULL) {
		RETVAL_NULL();
		return;
	}

	RETVAL_STRING(res, 1);
	free(res);
}

void PHP_RegisterCommand(void *client_data, const char *name)
{
	PHP_ClientData *PHP_MyData = client_data;
	zend_function_entry **func_arr = (zend_function_entry **)&PHP_MyData->functions;
	zend_function_entry *new_func;

 /* We take 2 extra, 1 new, and 1 NULL */
	*func_arr = realloc(*func_arr, sizeof(zend_function_entry) * (PHP_MyData->func_count + 2));
	PHP_MyData->func_count++;

	/* Clear the last 2 items */
	new_func = &((*func_arr)[PHP_MyData->func_count - 1]);
	memset(new_func, 0, sizeof(zend_function_entry) * 2);

	/* Now we are going to fill the new item */
	new_func = &((*func_arr)[PHP_MyData->func_count - 1]);

	new_func->fname = strdup(name);
	new_func->handler = PHP_DynamicPackageCommand;

	/* And register the item */
	if (zend_register_functions(
#if PHP5
	NULL,
#endif
	new_func, CG(function_table), MODULE_PERSISTENT) == FAILURE)
		gpmi_error(0, "[PHP] Could not register '%s'", name);
}

int PHP_LoadScript(void *client_data, const char *filename)
{
	zend_file_handle file_handle;

	file_handle.type = ZEND_HANDLE_FILENAME;
	file_handle.filename = estrdup(filename);
	file_handle.free_filename = 1;
	file_handle.opened_path = NULL;

	zend_first_try {
		if (zend_execute_scripts(ZEND_REQUIRE, NULL, 1, &file_handle) != SUCCESS) {
			gpmi_error(0, "[PHP] Could not run '%s'\n", file_handle.filename);
			return -4;
		}
	} zend_catch {
		gpmi_error(0, "[PHP] Critcal: could not run '%s'\n", file_handle.filename);
		return -3;
	} zend_end_try();

	return 0;
}

void PHP_CallFunction(void *client_data, const char *function, int argc, const char **argv)
{
	char *f = strdup(function);
	zval *function_name;
	zval *retval = NULL;
	zval **arg_real = NULL;
	zval ***my_argv = NULL;
	int i;

	if (argc != 0) {
		arg_real = malloc(sizeof(zval *) * argc);
		my_argv = malloc(sizeof(zval **) * argc);

		for (i = 0; i < argc; i++) {
			MAKE_STD_ZVAL(arg_real[i]);
			ZVAL_STRINGL(arg_real[i], (char *)argv[i], strlen(argv[i]), 1);
			my_argv[i] = &arg_real[i];
		}
	}

	MAKE_STD_ZVAL(function_name);
	ZVAL_STRINGL(function_name, f, strlen(f), 1);

	zend_first_try {
		call_user_function_ex(CG(function_table), NULL, function_name, &retval, argc, my_argv, 0, NULL);
	} zend_catch {
	} zend_end_try();

	if (retval != NULL)
		zval_ptr_dtor(&retval);

	for (i = 0; i < argc; i++)
		zval_ptr_dtor(&arg_real[i]);

	zval_ptr_dtor(&function_name);

	free(f);
	if (argc != 0) {
		free(my_argv);
		free(arg_real);
	}
}

int PHP_Initialize(void **client_data)
{
	PHP_ClientData *MyData;
	static char *argv[2] = {"gpmi", NULL};

	if (php_module_loaded >= 1) {
		gpmi_error(0, "[PHP] This module can only have one active instance");
		return -1;
	}

	if (php_embed_init(1, argv) == FAILURE) {
		gpmi_error(0, "[PHP] Unable to launch php/embed");
		return -2;
	}

	MyData = malloc(sizeof(PHP_ClientData));
	memset(MyData, 0, sizeof(PHP_ClientData));

	*client_data = MyData;

	php_module_loaded++;

	/* register the gpmi "extension" */
	zend_startup_module(&php_gpmi_module_entry);

	php_set_ini_entry("include_path", "./", PHP_INI_STAGE_ACTIVATE);
	php_set_ini_entry("html_errors", "0", PHP_INI_STAGE_ACTIVATE);
	php_set_ini_entry("display_errors", "1", PHP_INI_STAGE_ACTIVATE);
	php_set_ini_entry("display_startup_errors", "1", PHP_INI_STAGE_ACTIVATE);
	php_set_ini_entry("max_execution_time", "0", PHP_INI_STAGE_ACTIVATE);

	return 0;
}

void PHP_Destroy(void **client_data)
{
	PHP_ClientData *PHP_MyData = *client_data;
	zend_function_entry **func_arr = (zend_function_entry **)&PHP_MyData->functions;
	int i;

	/* Somehow PHP4 and PHP5 don't like it when we register stuff
	 *   while a script is loaded. To prevent closing errors, we keep track of
	 *   those new functions, and we free them ourself before we close the PHP
	 *   wrapper. Now everyone is happy :)
	 */
	zend_unregister_functions(*func_arr, PHP_MyData->func_count, CG(function_table));

	for (i = 0; i < PHP_MyData->func_count; i++) {
		free((*func_arr)[i].fname);
	}
	free(PHP_MyData->functions);

	php_embed_shutdown();

	php_module_loaded--;
	free(*client_data);
}

char *PHP_GetExtension()
{
	return strdup(".php");
}
